@@ -1,5 +1,5 @@
 /*
-Copyright (C) 2005, 2006 John E. Davis
+Copyright (C) 2005, 2006, 2007 John E. Davis
 
 This file is part of the S-Lang Library.
 
@@ -35,7 +35,7 @@ USA.
 #include <signal.h>
 #include <slang.h>
 
-static char *Slsh_Version = "0.7.5-0";
+static char *Slsh_Version = "0.8.2-0";
 #define SLSHRC_FILE "slsh.rc"
 #include "slsh.h"
 
@@ -187,7 +187,7 @@ static void at_exit (SLang_Ref_Type *ref)
    AtExit_Hooks = a;
 }
 
-static void c_exit (int *code)
+static void c_exit (int status)
 {
    while (AtExit_Hooks != NULL)
      {
@@ -198,11 +198,26 @@ static void c_exit (int *code)
 	SLfree ((char *) AtExit_Hooks);
 	AtExit_Hooks = next;
      }
+
    if (SLang_get_error ())
      SLang_restart (1);
-   exit (*code);
+
+   exit (status);
 }
 
+static void exit_intrin (void)
+{
+   int status;
+   
+   if (SLang_Num_Function_Args == 0)
+     status = 0;
+   else if (-1 == SLang_pop_int (&status))
+     return;
+   
+   c_exit (status);
+}
+
+
 
 static void stat_mode_to_string (void)
 {
@@ -404,10 +419,8 @@ static int setup_paths (void)
 /* Create the Table that S-Lang requires */
 static SLang_Intrin_Fun_Type Intrinsics [] =
 {
-   MAKE_INTRINSIC_I("exit", c_exit, VOID_TYPE),
+   MAKE_INTRINSIC_0("exit", exit_intrin, VOID_TYPE),
    MAKE_INTRINSIC_1("atexit", at_exit, VOID_TYPE, SLANG_REF_TYPE),
-   MAKE_INTRINSIC_S("slsh_readline", slsh_readline_intrinsic, VOID_TYPE),
-   MAKE_INTRINSIC_S("slsh_readline_noecho", slsh_readline_noecho_intrinsic, VOID_TYPE),
    MAKE_INTRINSIC_0("stat_mode_to_string", stat_mode_to_string, VOID_TYPE),
    SLANG_END_INTRIN_FUN_TABLE
 };
@@ -419,6 +432,7 @@ static void usage (void)
 Usage: slsh [OPTIONS] [-|file [args...]]\n\
  --help           Print this help\n\
  --version        Show slsh version information\n\
+ -e string        Execute 'string' as S-Lang code\n\
  -g               Compile with debugging code, tracebacks, etc\n\
  -n               Don't load personal init file\n\
  --init file      Use this file instead of ~/%s\n",
@@ -426,8 +440,9 @@ Usage: slsh [OPTIONS] [-|file [args...]]\n\
    fprintf (stderr, "\
  --no-readline    Do not use readline\n\
  -i               Force interactive input\n\
- -t               Test mode.  If slsh_main exits, do not call it\n\
+ -t               Test mode.  If slsh_main exists, do not call it\n\
  -v               Show verbose loading messages\n\
+ -Dname           Define \"name\" as a preprocessor symbol\n\
 \n\
   Note: - and -i are mutually exclusive\n\
 \n\
@@ -440,16 +455,30 @@ Usage: slsh [OPTIONS] [-|file [args...]]\n\
    exit (1);
 }
 
-static void version (void)
+static void output_version (void)
 {
-   fprintf (stdout, "slsh version %s\n", Slsh_Version);
-   fprintf (stdout, "S-Lang Library Version: %s\n", SLang_Version_String);
+   fprintf (stdout, "slsh version %s; ", Slsh_Version);
+   fprintf (stdout, "S-Lang version: %s\n", SLang_Version_String);
    if (SLANG_VERSION != SLang_Version)
      {
 	fprintf (stdout, "\t** Note: This program was compiled against version %s.\n",
 		 SLANG_VERSION_STRING);
      }
+}
+
+static int output_copyright (void)
+{
+   output_version ();
+   fprintf (stdout, "Copyright (C) 2005-2007 John E. Davis <jed@jedsoft.org>\r\n");
+   fprintf (stdout, "This is free software with ABSOLUTELY NO WARRANTY.\r\n");
+   fprintf (stdout, "\n");
    
+   return 0;
+}
+
+static void version (void)
+{
+   output_version ();
    exit (0);
 }
 
@@ -462,20 +491,15 @@ int main (int argc, char **argv)
    int is_interactive = 0;
    int use_readline = 1;
    int test_mode = 0;
-
-   if (SLang_Version < SLANG_VERSION)
-     {
-	fprintf (stderr, "***Warning: Executable compiled against S-Lang %s but linked to %s\n",
-		 SLANG_VERSION_STRING, SLang_Version_String);
-	fflush (stderr);
-     }
+   char *exec_string = NULL;
 
    (void) SLutf8_enable (-1);
 
    if ((-1 == SLang_init_all ())
        || (-1 == SLang_init_array_extra ())
        || (-1 == SLang_init_import ()) /* dynamic linking */
-       || (-1 == SLadd_intrin_fun_table (Intrinsics, NULL)))
+       || (-1 == SLadd_intrin_fun_table (Intrinsics, NULL))
+       || (-1 == slsh_init_readline_intrinsics ()))
      {
 	fprintf(stderr, "Unable to initialize S-Lang.\n");
 	return 1;
@@ -525,10 +549,19 @@ int main (int argc, char **argv)
 	     continue;
 	  }
 
+	if ((0 == strcmp (argv[1], "-e"))
+	    && (argc > 2))
+	  {
+	     argc -= 2;
+	     argv += 2;
+	     exec_string = *argv;
+	     continue;
+	  }
+
 	if (0 == strcmp (argv[1], "-g"))
 	  {
 	     SLang_generate_debug_info (1);
-	     SLang_Traceback = 1;
+	     SLang_Traceback = SL_TB_FULL;
 	     argc--;
 	     argv++;
 	     continue;
@@ -576,12 +609,23 @@ int main (int argc, char **argv)
 	     argv += 2;
 	     continue;
 	  }
+	
+	if (0 == strncmp (argv[1], "-D", 2))
+	  {
+	     char *prep = argv[1] + 2;
+	     if (*prep != 0)
+	       (void) SLdefine_for_ifdef (prep);
+	     argc--;
+	     argv++;
+	  }
+
 	break;
      }
 
    if (argc == 1)
      {
-	is_interactive = (isatty (fileno(stdin)) && isatty (fileno(stdout)));
+	if ((exec_string == NULL) && (is_interactive == 0))
+	  is_interactive = (isatty (fileno(stdin)) && isatty (fileno(stdout)));
 	file = NULL;
      }
    else
@@ -597,8 +641,14 @@ int main (int argc, char **argv)
 	argc--;
 	argv++;
      }
-   
-   (void) slsh_use_readline (use_readline);
+
+   if ((is_interactive == 0)
+       && (SLang_Version < SLANG_VERSION))
+     {
+	fprintf (stderr, "***Warning: Executable compiled against S-Lang %s but linked to %s\n",
+		 SLANG_VERSION_STRING, SLang_Version_String);
+	fflush (stderr);
+     }
 
    /* fprintf (stdout, "slsh: argv[0]=%s\n", argv[0]); */
    if (-1 == SLang_set_argc_argv (argc, argv))
@@ -610,12 +660,18 @@ int main (int argc, char **argv)
    if (-1 == load_startup_file (is_interactive))
      return SLang_get_error ();
 
+   /* Initializing the readline interface causes the .slrlinerc file
+    * to be loaded.  It is put here after the startup files have been loaded.
+    */
+   if (-1 == slsh_use_readline (SLpath_basename (argv[0]), use_readline, is_interactive))
+     return 1;
+
    if ((init_file != NULL)
        && (-1 == try_to_load_file (init_file_dir, init_file, NULL)))
      return SLang_get_error ();
 
-   /* Now load an initialization file and exit */
-   if ((file != NULL) || (is_interactive == 0))
+   if ((file != NULL)
+       || ((is_interactive == 0) && (exec_string == NULL)))
      {
 	if (0 == try_to_load_file (NULL, file, NULL))
 	  {
@@ -626,10 +682,21 @@ int main (int argc, char **argv)
 	  (void) SLang_run_hooks ("slsh_main", 0);
      }
 
+   if (exec_string != NULL)
+     {
+	(void) SLang_load_string (exec_string);
+     }
+
    if (is_interactive)
-     (void) slsh_interactive ();
+     {
+	output_copyright ();
+	if (SLang_Traceback != SL_TB_FULL)
+	  SLang_Traceback = SL_TB_NONE;
+
+	(void) slsh_interactive ();
+     }
 
    exit_val = SLang_get_error ();
-   c_exit (&exit_val);
+   c_exit (exit_val);
    return SLang_get_error ();
 }
