@@ -1,6 +1,6 @@
 /* User defined objects */
 /*
-Copyright (C) 2004, 2005, 2006 John E. Davis
+Copyright (C) 2004, 2005, 2006, 2007 John E. Davis
 
 This file is part of the S-Lang Library.
 
@@ -179,7 +179,8 @@ int _pSLclass_copy_class (SLtype to, SLtype from)
    Class_Table_Type *t;
 
    cl = _pSLclass_get_class (from);
-   clp = alloc_class_slot (to, &t);
+   if (NULL == (clp = alloc_class_slot (to, &t)))
+     return -1;
    
    if (*clp != NULL)
      {
@@ -1574,6 +1575,19 @@ SLclass_add_typecast (SLtype from, SLtype to,
 SLang_MMT_Type *SLang_pop_mmt (SLtype type) /*{{{*/
 {
    SLang_MMT_Type *mmt;
+   SLang_Class_Type *cl;
+   
+   cl = lookup_class (type);
+   if (cl == NULL)
+     {
+	SLang_verror (SL_Application_Error, "SLtype %d is not registered", type);
+	return NULL;
+     }
+   if (cl->cl_class_type != SLANG_CLASS_TYPE_MMT)
+     {
+	SLang_verror (SL_Application_Error, "SLtype %d is not an MMT", type);
+	return NULL;
+     }
 
    if (-1 == SLclass_pop_ptr_obj (type, VOID_STAR_STAR(&mmt)))
      mmt = NULL;
@@ -1859,3 +1873,40 @@ int SLang_get_int_size (SLtype type)
      }
    return 0;
 }
+
+int SLclass_patch_intrin_fun_table (SLang_Intrin_Fun_Type *table, 
+				  SLtype *from_types, SLtype *to_types, unsigned int n)
+{
+   unsigned int i, j;
+   
+   for (i = 0; i < n; i++)
+     {
+	SLang_Intrin_Fun_Type *t = table;
+	SLtype dummy = from_types[i];
+	SLtype type = to_types[i];
+
+	while (t->name != NULL)
+	  {
+	     unsigned int nargs = t->num_args;
+	     SLtype *args = t->arg_types;
+
+	     for (j = 0; j < nargs; j++)
+	       {
+		  if (args[j] == dummy)
+		    args[j] = type;
+	       }
+
+	     /* For completeness */
+	     if (t->return_type == dummy)
+	       t->return_type = type;
+	     t++;
+	  }
+     }
+   return 0;
+}
+int SLclass_patch_intrin_fun_table1 (SLang_Intrin_Fun_Type *table, 
+				   SLtype from_type, SLtype to_type)
+{
+   return SLclass_patch_intrin_fun_table (table, &from_type, &to_type, 1);
+}
+
