@@ -1,6 +1,6 @@
 /* interpreter signal handling functions */
 /*
-Copyright (C) 2004, 2005, 2006 John E. Davis
+Copyright (C) 2004, 2005, 2006, 2007 John E. Davis
 
 This file is part of the S-Lang Library.
 
@@ -235,7 +235,9 @@ static void signal_handler (int sig)
 
    s = find_signal (sig);	       /* cannot fail here */
 
-   s->pending = 1;   
+   s->pending = 1;
+   if (sig == SIGINT)
+     SLKeyBoard_Quit = 1;
    _pSLang_signal_interrupt ();
 }
 
@@ -458,7 +460,7 @@ static void signal_intrinsic (void)
 	return;
      }
 
-   old_handler = SLsignal (s->sig, signal_handler);
+   old_handler = SLsignal_intr (s->sig, signal_handler);
    if (-1 == set_old_handler (s, old_ref, old_handler))
      {
 	SLang_free_ref (old_ref);
