@@ -2,7 +2,7 @@
 /* Standard intrinsic functions for S-Lang.  Included here are string
    and array operations */
 /*
-Copyright (C) 2004, 2005 John E. Davis
+Copyright (C) 2004, 2005, 2006, 2007 John E. Davis
 
 This file is part of the S-Lang Library.
 
@@ -102,8 +102,16 @@ static int length_cmd (void)
 static void char_cmd (SLwchar_Type *x) /*{{{*/
 {
    SLuchar_Type buf[SLUTF8_MAX_MBLEN + 1];
-
-   if ((_pSLinterp_UTF8_Mode == 0) || (*x < 0x80))
+   int is_byte;
+   
+   is_byte = ((signed)*x < 0);
+   if (is_byte)
+     {
+	buf[0] = (SLuchar_Type) (-(signed)*x);
+	buf[1] = 0;
+     }
+   else if ((_pSLinterp_UTF8_Mode == 0) 
+	    || (*x < 0x80))
      {
         buf[0] = (SLuchar_Type) *x;
         buf[1] = 0;
@@ -607,7 +615,7 @@ static int intrin_get_defines (void)
 
 static void intrin_get_reference (char *name)
 {
-   _pSLang_push_ref (1, (VOID_STAR) _pSLlocate_name (name));
+   _pSLang_push_nt_as_ref (_pSLlocate_name (name));
 }
 
 static void intrin_get_namespaces (void)
@@ -731,12 +739,17 @@ static int is_numeric (SLtype type)
    /* Version 2: Add attributes to the class tables to simplify this.  
     * Also clarify exactly what _pSLang_is_arith_type is supposed to return.
     */
-   if (0 != _pSLang_is_arith_type ((SLtype) type))
-     return 1;
-   if (type == SLANG_COMPLEX_TYPE)
-     return 1;
+   if (0 == _pSLang_is_arith_type ((SLtype) type))
+     {
+	if (type == SLANG_COMPLEX_TYPE)
+	  return 3;
+	
+	return 0;
+     }
+   if ((type == SLANG_DOUBLE_TYPE) || (type == SLANG_FLOAT_TYPE))
+     return 2;
 
-   return 0;
+   return 1;
 }
 
 static int is_numeric_intrinsic (void)
@@ -805,6 +818,78 @@ static int stack_depth_intrin (void)
    return SLstack_depth ();
 }
 
+static void expand_dollar_string (char *s)
+{
+   (void) _pSLpush_dollar_string (s);
+}
+
+#if SLANG_HAS_QUALIFIERS
+static void get_qualifiers_intrin (void)
+{
+   SLang_Struct_Type *q;
+   if (0 == _pSLang_get_qualifiers (&q))
+     (void) SLang_push_struct (q);
+}
+
+static int qualifier_exists_intrin (char *name)
+{
+   SLang_Struct_Type *q;
+
+   if (-1 == _pSLang_get_qualifiers (&q))
+     return -1;
+   
+   if ((q == NULL)
+       || (NULL == _pSLstruct_get_field_value (q, name)))
+     return 0;
+   
+   return 1;
+}
+
+static void qualifier_intrin (void)
+{
+   int has_default;
+   char *name;
+   SLang_Struct_Type *q;
+   SLang_Object_Type *objp;
+
+   if (-1 == _pSLang_get_qualifiers (&q))
+     return;
+
+   has_default = (SLang_Num_Function_Args == 2);
+   if (has_default)
+     {
+	if (-1 == SLroll_stack (2))
+	  return;
+     }
+
+   if (-1 == SLang_pop_slstring (&name))
+     return;
+
+   if (q != NULL)
+     objp = _pSLstruct_get_field_value (q, name);
+   else
+     objp = NULL;
+   
+   SLang_free_slstring (name);
+
+   if (objp != NULL)
+     {
+	if (has_default)
+	  SLdo_pop ();
+	_pSLpush_slang_obj (objp);
+     }
+   else if (has_default == 0)
+     (void) SLang_push_null ();
+   
+   /* Note: objp and q should _not_ be freed since they were not allocated */
+}
+#endif
+
+static void clear_error_intrin (void)
+{
+   (void) _pSLerr_clear_error (1);
+}
+
 static void set_argv_intrinsic (void);
 static SLang_Intrin_Fun_Type SLang_Basic_Table [] = /*{{{*/
 {
@@ -858,7 +943,7 @@ static SLang_Intrin_Fun_Type SLang_Basic_Table [] = /*{{{*/
    MAKE_INTRINSIC_0("_print_stack", lang_print_stack, SLANG_VOID_TYPE),
    MAKE_INTRINSIC_I("_stk_roll", intrin_roll_stack, SLANG_VOID_TYPE),
    MAKE_INTRINSIC_SI("byte_compile_file", byte_compile_file, SLANG_VOID_TYPE),
-   MAKE_INTRINSIC_0("_clear_error", _pSLerr_clear_error, SLANG_VOID_TYPE),
+   MAKE_INTRINSIC_0("_clear_error", clear_error_intrin, SLANG_VOID_TYPE),
    MAKE_INTRINSIC_0("_function_name", intrin_function_name, SLANG_STRING_TYPE),
 #if SLANG_HAS_FLOAT
    MAKE_INTRINSIC_S("set_float_format", _pSLset_double_format, SLANG_VOID_TYPE),
@@ -875,6 +960,12 @@ static SLang_Intrin_Fun_Type SLang_Basic_Table [] = /*{{{*/
    MAKE_INTRINSIC_0("current_namespace", _pSLang_cur_namespace_intrinsic, SLANG_STRING_TYPE),
    MAKE_INTRINSIC_0("length", length_cmd, SLANG_INT_TYPE),
    MAKE_INTRINSIC_0("__set_argc_argv", set_argv_intrinsic, SLANG_VOID_TYPE),
+   MAKE_INTRINSIC_S("_$", expand_dollar_string, SLANG_VOID_TYPE),
+#if SLANG_HAS_QUALIFIERS
+   MAKE_INTRINSIC_0("__qualifiers", get_qualifiers_intrin, SLANG_VOID_TYPE),
+   MAKE_INTRINSIC_0("qualifier", qualifier_intrin, SLANG_VOID_TYPE),
+   MAKE_INTRINSIC_S("qualifier_exists", qualifier_exists_intrin, SLANG_INT_TYPE),
+#endif
    SLANG_END_INTRIN_FUN_TABLE
 };
 
@@ -892,11 +983,13 @@ static char *Install_Prefix = SLANG_INSTALL_PREFIX;
 static char *Install_Prefix = "";
 #endif
 
+static int obsolete_int_variable;
 static SLang_Intrin_Var_Type Intrin_Vars[] =
 {
-   MAKE_VARIABLE("_debug_info", &_pSLang_Compile_Line_Num_Info, SLANG_INT_TYPE, 0),
+   MAKE_VARIABLE("_debug_info", &obsolete_int_variable, SLANG_INT_TYPE, 0),
 #if SLANG_HAS_BOSEOS
    MAKE_VARIABLE("_boseos_info", &_pSLang_Compile_BOSEOS, SLANG_INT_TYPE, 0),
+   MAKE_VARIABLE("_bofeof_info", &_pSLang_Compile_BOFEOF, SLANG_INT_TYPE, 0),
 #endif
    MAKE_VARIABLE("_auto_declare", &_pSLang_Auto_Declare_Globals, SLANG_INT_TYPE, 0),
    MAKE_VARIABLE("_slangtrace", &_pSLang_Trace, SLANG_INT_TYPE, 0),
